#ifndef DUNE_TECTONIC_LINEARIZATION_HH
#define DUNE_TECTONIC_LINEARIZATION_HH

#include <dune/tnnmg/functionals/bcqfconstrainedlinearization.hh>
#include <dune/istl/matrixindexset.hh>

//#include <>

/*#include <dune/common/bitsetvector.hh>
#include <dune/common/parametertree.hh>
#include <dune/common/fmatrixev.hh>

#include <dune/fufem/arithmetic.hh>
#include <dune/solvers/common/interval.hh>
#include <dune/solvers/computeenergy.hh>
#include <dune/tnnmg/problem-classes/bisection.hh>
#include <dune/tnnmg/problem-classes/blocknonlineargsproblem.hh>

#include <dune/tectonic/globalfriction.hh>
#include <dune/tectonic/minimisation.hh>
#include <dune/tectonic/mydirectionalconvexfunction.hh>
#include <dune/tectonic/quadraticenergy.hh>*/


template<class F, class BV>
class Linearization : public Dune::TNNMG::BoxConstrainedQuadraticFunctionalConstrainedLinearization<F,BV> {
private:
  using Base = Dune::TNNMG::BoxConstrainedQuadraticFunctionalConstrainedLinearization<F,BV>;
  using Vector = typename Base::Vector;
  using BitVector = typename Base::BitVector;

  template <class T>
  void determineRegularityTruncation(const Vector& x, BitVector&& truncationFlags, const T& truncationTolerance)
  {
    for (size_t i = 0; i < x.size(); ++i) {
      if (this->f_.phi().regularity(i, x[i]) > truncationTolerance) {
        truncationFlags[i] = true;
      }
    }
  }

public:
  Linearization(const F& f, const BitVector& ignore) :
    Base(f, ignore)
  {}

  void bind(const Vector& x) {
      auto&& A = this->f_.quadraticPart();
      auto&& phi = this->f_.phi();

      // compute hessian
      // ---------------

      // construct sparsity pattern for linearization
      Dune::MatrixIndexSet indices(A.N(), A.M());
      indices.import(A);
      phi.addHessianIndices(indices);

      // construct matrix from pattern and initialize it
      indices.exportIdx(this->hessian_);
      this->hessian_ = 0.0;

      // quadratic part
      for (size_t i = 0; i < A.N(); ++i) {
        auto const end = std::end(A[i]);
        for (auto it = std::begin(A[i]); it != end; ++it)
          this->hessian_[i][it.index()] += *it;
      }

      // nonlinearity part
      phi.addHessian(x, this->hessian_);

      // compute gradient
      // ----------------

      // quadratic part
      this->negativeGradient_ = derivative(this->f_)(x); // A*x - b

      // nonlinearity part
      phi.addGradient(x, this->negativeGradient_);

      // -grad is needed for Newton step
      this->negativeGradient_ *= -1;


      // determine which components to truncate
      // ----------------
      this->truncationFlags_ = this->ignore_;

      determineTruncation(x, this->f_.lowerObstacle(), this->f_.upperObstacle(), this->truncationFlags_, this->truncationTolerance_); // obstacle truncation
      determineRegularityTruncation(x, this->truncationFlags_, this->truncationTolerance_); // truncation due to regularity deficit of nonlinearity

      // truncate gradient and hessian
      truncateVector(this->negativeGradient_, this->truncationFlags_);
      truncateMatrix(this->hessian_, this->truncationFlags_, this->truncationFlags_);
  }
};

#endif
